# MIT License

# Copyright (c) 2022 Cheng

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# GIN Code based on:
# Ouyang, C., Chen, C., Li, S., Li, Z., Qin, C., Bai, W., & Rueckert, D. (2022). Causality-inspired single-source domain generalization for medical image segmentation. IEEE Transactions on Medical Imaging, 42(4), 1095-1106.
# DOI: 10.1109/TMI.2022.3224067
# https://github.com/cheng-01037/Causality-Medical-Image-Domain-Generalization
# https://ieeexplore.ieee.org/document/9961940

import torch
import torch.nn as nn
import torch.nn.functional as F

from dg_tta.utils import get_internal_augmentation_enabled


class GradlessGCReplayNonlinBlock(nn.Module):
    def __init__(
        self,
        out_channel=32,
        in_channel=3,
        scale_pool=[1, 3],
        layer_id=0,
        use_act=True,
        requires_grad=False,
    ):
        """
        Conv-leaky relu layer. Efficient implementation by using group convolutions
        """
        super(GradlessGCReplayNonlinBlock, self).__init__()

        self.in_channel = in_channel
        self.out_channel = out_channel
        self.scale_pool = scale_pool
        self.layer_id = layer_id
        self.use_act = use_act
        self.requires_grad = requires_grad
        assert requires_grad == False

    def forward(self, x_in, requires_grad=False):
        """
        Args:
            x_in: [ nb (original), nc (original), nx, ny ]
        """
        # random size of kernel
        idx_k = torch.randint(high=len(self.scale_pool), size=(1,))
        k = self.scale_pool[idx_k[0]]

        if x_in.dim() == 5:
            mode = "3d"
        elif x_in.dim() == 4:
            mode = "2d"
        else:
            raise ValueError()

        if mode == "2d":
            nb, nc, nx, ny = x_in.shape
            ker = torch.randn(
                [self.out_channel * nb, self.in_channel, k, k],
                requires_grad=self.requires_grad,
            ).to(x_in.device)
            shift = (
                torch.randn(
                    [self.out_channel * nb, 1, 1], requires_grad=self.requires_grad
                ).to(x_in.device)
                * 1.0
            )
            x_in = x_in.reshape(1, nb * nc, nx, ny)
            x_conv = F.conv2d(
                x_in, ker, stride=1, padding=k // 2, dilation=1, groups=nb
            )

        elif mode == "3d":
            nb, nc, nx, ny, nz = x_in.shape
            ker = torch.randn(
                [self.out_channel * nb, self.in_channel, k, k, k],
                requires_grad=self.requires_grad,
            ).to(x_in.device)
            shift = (
                torch.randn(
                    [self.out_channel * nb, 1, 1, 1], requires_grad=self.requires_grad
                ).to(x_in.device)
                * 1.0
            )
            x_in = x_in.reshape(1, nb * nc, nx, ny, nz)
            x_conv = F.conv3d(
                x_in, ker, stride=1, padding=k // 2, dilation=1, groups=nb
            )
        else:
            raise ValueError()

        x_conv = x_conv + shift
        if self.use_act:
            x_conv = F.leaky_relu(x_conv)

        if mode == "2d":
            x_conv = x_conv.reshape(nb, self.out_channel, nx, ny)
        elif mode == "3d":
            x_conv = x_conv.reshape(nb, self.out_channel, nx, ny, nz)
        else:
            raise ValueError()

        return x_conv


class GINGroupConv(nn.Module):
    def __init__(self, cfg):
        """
        GIN
        """
        super(GINGroupConv, self).__init__()
        self.scale_pool = [1, 3]  # don't make it tool large as we have multiple layers
        self.n_layer = cfg["N_LAYER"]
        self.layers = []
        self.out_norm = "frob"
        self.out_channel = cfg["IN_CHANNELS"]
        in_channel = cfg["IN_CHANNELS"]
        interm_channel = cfg["INTERM_CHANNELS"]

        self.layers.append(
            GradlessGCReplayNonlinBlock(
                out_channel=interm_channel,
                in_channel=in_channel,
                scale_pool=self.scale_pool,
                layer_id=0,
            )
        )
        for ii in range(self.n_layer - 2):
            self.layers.append(
                GradlessGCReplayNonlinBlock(
                    out_channel=interm_channel,
                    in_channel=interm_channel,
                    scale_pool=self.scale_pool,
                    layer_id=ii + 1,
                )
            )
        self.layers.append(
            GradlessGCReplayNonlinBlock(
                out_channel=self.out_channel,
                in_channel=interm_channel,
                scale_pool=self.scale_pool,
                layer_id=self.n_layer - 1,
                use_act=False,
            )
        )

        self.layers = nn.ModuleList(self.layers)

    def forward(self, x_in):
        if isinstance(x_in, list):
            x_in = torch.cat(x_in, dim=0)

        if x_in.dim() == 5:
            mode = "3d"
        elif x_in.dim() == 4:
            mode = "2d"
        else:
            raise ValueError()

        nb, nc = x_in.shape[:2]

        if mode == "2d":
            alphas = torch.rand(nb, device=x_in.device)[
                :, None, None, None
            ]  # nb, 1, 1, 1
            alphas = alphas.repeat(1, nc, 1, 1)  # nb, nc, 1, 1
        elif mode == "3d":
            alphas = torch.rand(nb, device=x_in.device)[
                :, None, None, None, None
            ]  # nb, 1, 1, 1, 1
            alphas = alphas.repeat(1, nc, 1, 1, 1)  # nb, nc, 1, 1, 1
        else:
            raise ValueError()

        x = self.layers[0](x_in)
        for blk in self.layers[1:]:
            x = blk(x)
        mixed = alphas * x + (1.0 - alphas) * x_in

        if self.out_norm == "frob":
            _in_frob = torch.norm(
                x_in.reshape(nb, nc, -1), dim=(-1, -2), p="fro", keepdim=False
            )
            if mode == "2d":
                _in_frob = _in_frob[:, None, None, None].repeat(1, nc, 1, 1)
                _self_frob = torch.norm(
                    mixed.reshape(nb, self.out_channel, -1),
                    dim=(-1, -2),
                    p="fro",
                    keepdim=False,
                )
                _self_frob = _self_frob[:, None, None, None].repeat(
                    1, self.out_channel, 1, 1
                )
            elif mode == "3d":
                _in_frob = _in_frob[:, None, None, None, None].repeat(1, nc, 1, 1, 1)
                _self_frob = torch.norm(
                    mixed.reshape(nb, self.out_channel, -1),
                    dim=(-1, -2),
                    p="fro",
                    keepdim=False,
                )
                _self_frob = _self_frob[:, None, None, None, None].repeat(
                    1, self.out_channel, 1, 1, 1
                )
            else:
                raise ValueError()

            mixed = mixed * (1.0 / (_self_frob + 1e-5)) * _in_frob

        return mixed


def gin_aug(input):
    cfg = dict(
        IN_CHANNELS=1,
        N_LAYER=4,
        INTERM_CHANNELS=2,
    )
    gin_group_conv = GINGroupConv(cfg)
    input = gin_group_conv(input)
    return input


def gin_hook(module, input):
    if get_internal_augmentation_enabled():
        return gin_aug(*input)
    return input



from collections import OrderedDict

import torch
import torch.nn.functional as F

MOD_GET_FN = (
    lambda self, key: self[int(key)]
    if isinstance(self, torch.nn.Sequential)
    else getattr(self, key)
)


def get_batch(tensor_list, batch_idx, patch_size, fixed_patch_idx=None, device="cpu"):
    assert (
        fixed_patch_idx in range(8)
        or fixed_patch_idx == None
        or fixed_patch_idx == "center"
    )

    device = torch.device(device)
    B = len(batch_idx)
    b_img = []
    b_label = []

    for b in range(B):
        with torch.no_grad():
            # Get patches
            data = tensor_list[batch_idx[b]]
            if fixed_patch_idx is None:
                rand_patch_d = torch.randint(
                    max(data.shape[1] - patch_size[0], 0), (1,)
                )
                rand_patch_h = torch.randint(
                    max(data.shape[2] - patch_size[1], 0), (1,)
                )
                rand_patch_w = torch.randint(
                    max(data.shape[3] - patch_size[2], 0), (1,)
                )

            elif fixed_patch_idx == "center":
                rand_patch_d = max((data.shape[1] - patch_size[0]) // 2, 0)
                rand_patch_h = max((data.shape[2] - patch_size[1]) // 2, 0)
                rand_patch_w = max((data.shape[3] - patch_size[2]) // 2, 0)

            else:
                p_idxs = f"{fixed_patch_idx:03b}"
                p_idxs = [int(idx) for idx in [*p_idxs]]
                rand_patch_d = p_idxs[0] * patch_size[0]
                rand_patch_h = p_idxs[1] * patch_size[1]
                rand_patch_w = p_idxs[2] * patch_size[2]
            # print(rand_patch_d, rand_patch_h, rand_patch_w)

            out_shape = (
                1,
                1,
                max(data.shape[1], patch_size[0]),
                max(data.shape[2], patch_size[1]),
                max(data.shape[3], patch_size[2]),
            )
            grid = F.affine_grid(
                torch.eye(3, 4).unsqueeze(0).to(device), out_shape, align_corners=False
            )
            patch_grid = grid[
                :,
                rand_patch_d : rand_patch_d + patch_size[0],
                rand_patch_h : rand_patch_h + patch_size[1],
                rand_patch_w : rand_patch_w + patch_size[2],
            ]

            b_img.append(F.grid_sample(
                data[0:1].unsqueeze(0).to(device), patch_grid, align_corners=False
            ))

            if data[1:].numel() == 0:
                # No GT label is available for this sample
                b_label.append(None)
            else:
                b_label.append(
                    get_argmaxed_segs(
                        F.grid_sample(
                            data[1:].unsqueeze(0).to(device),
                            patch_grid,
                            align_corners=False,
                            mode="nearest",
                        )
                    )
                )

    return b_img, b_label


def get_argmaxed_segs(segs):
    segs_oh_w_bg = torch.cat([(segs.sum(1, keepdim=True) < 1.0).float(), segs], dim=1)
    segs_argmaxed = segs_oh_w_bg.argmax(1, keepdim=True)
    return segs_argmaxed


def get_imgs(tta_sample):
    imgs = tta_sample[:, 0:1]
    return imgs


def soft_dice_loss(smp_a, smp_b):
    B, _, D, H, W = smp_a.shape
    d = 2

    nominator = (2.0 * smp_a * smp_b).reshape(B, -1, D * H * W).mean(2)
    denominator = 1 / d * ((smp_a + smp_b) ** d).reshape(B, -1, D * H * W).mean(2)

    if denominator.sum() == 0.0:
        dice = (nominator * 0.0) + 1.0
    else:
        dice = (
            nominator / denominator
        )  # Do not add an eps here, it disturbs the consistency

    return dice


def dice_coeff(outputs, labels, max_label):
    dice = torch.FloatTensor(max_label - 1).fill_(0)

    for label_num in range(1, max_label):
        iflat = (outputs == label_num).view(-1).float()
        tflat = (labels == label_num).view(-1).float()
        intersection = torch.mean(iflat * tflat)
        dice[label_num - 1] = (2.0 * intersection) / (
            1e-8 + torch.mean(iflat) + torch.mean(tflat)
        )
    return dice


def fix_all(m):
    for p in m.parameters():
        p.requires_grad_(False)


def release_all(m):
    for p in m.parameters():
        p.requires_grad_(True)


def release_norms(m):
    if (
        "instancenorm" in m.__class__.__name__.lower()
        or "batchnorm" in m.__class__.__name__.lower()
    ):
        print("Released", m.__class__.__name__)
        for p in m.parameters():
            p.requires_grad_(True)


import functools


def get_named_layers_leaves(module):
    """Returns all leaf layers of a pytorch module and a keychain as identifier.
    e.g.
    ...
    ('features.0.5', nn.ReLU())
    ...
    ('classifier.0', nn.BatchNorm2D())
    ('classifier.1', nn.Linear())
    """

    return [
        (keychain, sub_mod)
        for keychain, sub_mod in list(module.named_modules())
        if not next(sub_mod.children(), None)
    ]


def get_module_data(module, keychain):
    """Retrieves any data inside a pytorch module for a given keychain.
    Use get_named_layers_leaves(module) to retrieve valid keychains for layers.
    """

    return functools.reduce(MOD_GET_FN, keychain.split("."), module)


def set_module_data(module, keychain, data):
    """Replaces any data inside a pytorch module for a given keychain.
    Use get_named_layers_leaves(module) to retrieve valid keychains for layers.
    """
    key_list = keychain.split(".")
    root = functools.reduce(MOD_GET_FN, key_list[:-1], module)
    leaf = key_list[-1]
    if isinstance(root, torch.nn.Sequential):
        root[int(leaf)] = data
    else:
        setattr(root, leaf, data)


def register_forward_pre_hook_at_beginning(model, hook_fn):
    hook_dict = {
        k: v
        for k, v in zip(
            range(len(model._forward_pre_hooks) + 1),
            [hook_fn] + list(model._forward_pre_hooks.values()),
        )
    }
    model._forward_pre_hooks = OrderedDict(hook_dict)


def register_forward_hook_at_beginning(model, hook_fn):
    hook_dict = {
        k: v
        for k, v in zip(
            range(len(model._forward_hooks) + 1),
            [hook_fn] + list(model._forward_hooks.values()),
        )
    }
    model._forward_hooks = OrderedDict(hook_dict)


def hookify(fn, type):
    assert type in ["forward_pre_hook", "forward_hook"]

    if type == "forward_pre_hook":
        return lambda module, input: fn(*input)
    elif type == "forward_hook":
        return lambda module, input, output: fn(output)

    raise ValueError()


def map_label(label, map_idxs, input_format):
    assert input_format in ["logits", "argmaxed"]

    if input_format == "logits":
        # We have a non argmaxed map, suppose that C dimension is label dimension
        mapped_label = label
        # Swap B,C and subselect
        mapped_label = mapped_label.transpose(0, 1)[map_idxs].transpose(0, 1)
    else:
        mapped_label = torch.zeros_like(label)
        for lbl_idx, map_idx in enumerate(map_idxs):
            mapped_label[label == map_idx] = lbl_idx

    return mapped_label


def generate_label_mapping(source_label_dict, target_label_dict):
    assert all([isinstance(k, str) for k in source_label_dict.keys()])
    assert all([isinstance(k, str) for k in target_label_dict.keys()])
    assert set(source_label_dict.keys()).intersection(
        target_label_dict.keys()
    ), "There are no intersecting label names in given dicts."
    mapped_label = []

    mapping_dict = dict.fromkeys(
        list(source_label_dict.keys()) + list(target_label_dict.keys())
    )

    for key in mapping_dict:
        if key in source_label_dict and key in target_label_dict:
            mapping_dict[key] = (source_label_dict[key], target_label_dict[key])

    return {k: v for k, v in mapping_dict.items() if v is not None}


def get_map_idxs(label_mapping: dict, optimized_labels: list, input_type):
    assert input_type in ["pretrain_labels", "tta_labels"]
    assert optimized_labels[0] == "background"

    # Generate idxs from label_mapping dict
    map_idxs_list = []
    for reduced_idx, eval_label in enumerate(optimized_labels):
        src_idx, target_idx = label_mapping[eval_label]
        # map_idxs_list = [tts_dict[k] for k,v in amos_bcv_dict.items()]
        append_idx = src_idx if input_type == "pretrain_labels" else target_idx
        map_idxs_list.append(append_idx)

    map_idxs = torch.as_tensor(map_idxs_list)

    return map_idxs



from itertools import tee
from pathlib import Path
import importlib
import shutil
import json
import json

from contextlib import nullcontext

import torch
import torch.nn.functional as F

from tqdm import trange, tqdm

if importlib.util.find_spec("wandb"):
    import wandb

from nnunetv2.evaluation.evaluate_predictions import compute_metrics_on_folder_simple
from nnunetv2.imageio.simpleitk_reader_writer import SimpleITKIO

from dg_tta.utils import disable_internal_augmentation
from dg_tta.gin import gin_aug
from dg_tta.tta.torch_utils import (
    get_batch,
    map_label,
    dice_coeff,
    soft_dice_loss,
    fix_all,
    release_all,
    release_norms,
    get_map_idxs,
    get_imgs,
)
from dg_tta.tta.augmentation_utils import get_disp_field, get_rand_affine
from dg_tta.tta.config_log_utils import (
    get_global_idx,
    wandb_run_is_available,
    plot_run_results,
    get_parameters_save_path,
)
from dg_tta.tta.model_utils import (
    get_model_from_network,
    buffer_running_stats,
    apply_running_stats,
)
from dg_tta.tta.nnunet_utils import (
    run_inference,
    load_network,
    load_tta_data,
)

INTENSITY_AUG_FUNCTION_DICT = {"disabled": lambda img: img, "GIN": gin_aug}


def get_sample_specs(smp_idx, tta_data, save_path, across_all_samples=False):
    if across_all_samples:
        tta_sample = None
        tta_tens_list = [e["data"] for e in tta_data]
        sample_id = "all_samples"
        sub_dir_tta = save_path / "tta_output"
    else:
        tta_sample = next(tta_data)
        # tta_sample = tta_data[smp_idx] # In case you don't use a generator
        tta_tens_list = [tta_sample["data"]]
        sample_id = tta_sample["ofile"]
        sub_dir_tta = save_path / Path(sample_id).parent

    return tta_sample, tta_tens_list, sample_id, sub_dir_tta


def tta_main(
    run_name,
    config,
    tta_data_dir,
    save_base_path,
    label_mapping,
    modifier_fn_module,
    device,
    debug=False,
):
    START_CLASS = 1  # Do not use background for consistency loss

    # Load model
    pretrained_weights_filepath = config["pretrained_weights_filepath"]
    predictor, patch_size, network, parameters = load_network(
        pretrained_weights_filepath, device
    )

    # Load TTA data
    tta_across_all_samples = config["tta_across_all_samples"]

    tqdm.write("\n# Loading data")
    tta_data, num_samples = load_tta_data(
        config, tta_data_dir, predictor, tta_across_all_samples
    )

    if tta_across_all_samples:
        # TTA data is a list
        inference_data = tta_data
    else:
        # TTA data is a generator, so we need to copy it for inference
        tta_data, inference_data = tee(tta_data)

    ensemble_count = config["ensemble_count"]
    B = config["batch_size"]
    patches_to_be_accumulated = config["patches_to_be_accumulated"]
    tta_eval_patches = config["tta_eval_patches"]
    num_epochs = config["epochs"]
    start_tta_at_epoch = config["start_tta_at_epoch"]

    optimized_labels = config["optimized_labels"]

    save_path = Path(save_base_path) / run_name
    save_path.mkdir(exist_ok=True, parents=False)

    with open(save_path / "tta_plan.json", "w") as f:
        json.dump({k: v for k, v in config.items()}, f, indent=4)

    sitk_io = SimpleITKIO()

    identity_grid = F.affine_grid(
        torch.eye(4, device=device).repeat(B, 1, 1)[:, :3],
        [B, 1] + patch_size,
        align_corners=False,
    )

    if tta_across_all_samples:
        sample_range = [0]
    else:
        sample_range = trange(num_samples, desc="Samples")

    disable_internal_augmentation()

    tqdm.write("\n# Starting TTA")
    for smp_idx in sample_range:
        _, tta_tens_list, sample_id, sub_dir_tta = get_sample_specs(
            smp_idx, tta_data, save_path, tta_across_all_samples
        )
        tqdm.write(f"\nSample {sample_id}")

        sub_dir_tta.mkdir(exist_ok=True)

        for ensemble_idx in trange(ensemble_count, desc="Ensembles"):
            tta_parameters_save_path = get_parameters_save_path(
                sub_dir_tta, sample_id, ensemble_idx
            )
            if tta_parameters_save_path.is_file():
                tqdm.write(
                    f"TTA parameters file already exists. Skipping '{tta_parameters_save_path}'"
                )
                continue

            tta_losses = torch.zeros(num_epochs)
            eval_dices = torch.zeros(num_epochs)

            intensity_aug_func = INTENSITY_AUG_FUNCTION_DICT[
                config["intensity_aug_function"]
            ]

            model = get_model_from_network(network, modifier_fn_module, parameters)
            model = model.to(device)

            optimizer = torch.optim.AdamW(model.parameters(), lr=config["lr"])

            tbar = trange(num_epochs, desc="Epoch")

            model.apply(fix_all)
            for epoch in tbar:
                model.train()
                global_idx = get_global_idx(
                    [
                        (smp_idx, num_samples),
                        (ensemble_idx, ensemble_count),
                        (epoch, num_epochs),
                    ]
                )
                if wandb_run_is_available():
                    wandb.log({"ref_epoch_idx": epoch}, global_idx)
                step_losses = []

                if epoch == start_tta_at_epoch:
                    model.apply(fix_all)
                    if config["params_with_grad"] == "all":
                        model.apply(release_all)
                    elif config["params_with_grad"] == "norms":
                        model.apply(release_norms)
                    elif config["params_with_grad"] == "encoder":
                        model.encoder.apply(release_all)
                    else:
                        raise ValueError()

                    grad_params = {
                        id(p): p.numel() for p in model.parameters() if p.requires_grad
                    }
                    tqdm.write(
                        f"Released #{sum(list(grad_params.values()))/1e6:.2f} million trainable params"
                    )

                for _ in range(patches_to_be_accumulated):
                    with torch.no_grad():
                        imgs, _ = get_batch(
                            tta_tens_list,
                            torch.randperm(len(tta_tens_list))[:B],
                            patch_size,
                            fixed_patch_idx=None,
                            device=device,
                        )

                    imgs = torch.cat(imgs, dim=0)

                    target_a = calc_branch(
                        "branch_a",
                        config,
                        model,
                        intensity_aug_func,
                        identity_grid,
                        patch_size,
                        B,
                        label_mapping,
                        optimized_labels,
                        modifier_fn_module,
                        imgs,
                        device,
                    )
                    target_b = calc_branch(
                        "branch_b",
                        config,
                        model,
                        intensity_aug_func,
                        identity_grid,
                        patch_size,
                        B,
                        label_mapping,
                        optimized_labels,
                        modifier_fn_module,
                        imgs,
                        device,
                    )

                    # Apply consistency loss
                    common_content_mask = (
                        target_a.sum(1, keepdim=True) > 0.0
                    ).float() * (target_b.sum(1, keepdim=True) > 0.0).float()
                    sm_a = target_a.softmax(1) * common_content_mask
                    sm_b = target_b.softmax(1) * common_content_mask

                    loss = 1 - soft_dice_loss(sm_a, sm_b)[:, START_CLASS:].mean()

                    loss_accum = loss / patches_to_be_accumulated
                    step_losses.append(loss.detach().cpu())

                    if epoch >= start_tta_at_epoch:
                        loss_accum.backward()

                if epoch >= start_tta_at_epoch:
                    optimizer.step()
                    optimizer.zero_grad()

                tta_losses[epoch] = torch.stack(step_losses).mean().item()

                with torch.inference_mode():
                    model.eval()
                    for _ in range(tta_eval_patches):
                        imgs, labels = get_batch(
                            tta_tens_list,
                            torch.randperm(len(tta_tens_list))[:B],
                            patch_size,
                            fixed_patch_idx="center",  # This is just for evaluation purposes
                            device=device,
                        )

                        imgs = torch.cat(imgs, dim=0)

                        none_labels = [l is None for l in labels]
                        filtered_imgs = imgs[~torch.as_tensor(none_labels)]
                        filtered_labels = [
                            l for flag, l in zip(none_labels, labels) if not flag
                        ]

                        if len(filtered_imgs) == 0:
                            eval_dices[epoch] = float("nan")
                            continue

                        else:
                            filtered_labels = torch.cat(filtered_labels, dim=0)
                            output_eval = model(filtered_imgs)
                            if isinstance(output_eval, tuple):
                                output_eval = output_eval[0]

                            output_eval = map_label(
                                output_eval,
                                get_map_idxs(
                                    label_mapping,
                                    optimized_labels,
                                    input_type="pretrain_labels",
                                ),
                                input_format="logits",
                            )
                            target_argmax = output_eval.argmax(1)

                            filtered_labels = map_label(
                                filtered_labels,
                                get_map_idxs(
                                    label_mapping,
                                    optimized_labels,
                                    input_type="tta_labels",
                                ),
                                input_format="argmaxed",
                            ).long()
                            d_tgt_val = dice_coeff(
                                target_argmax, filtered_labels, len(optimized_labels)
                            )

                            eval_dices[epoch] += (
                                1 / tta_eval_patches * d_tgt_val.mean().item()
                            )

                    if debug:
                        break

                tbar.set_description(
                    f"Epochs, loss={tta_losses[epoch]:.3f}, Pseudo-Dice={eval_dices[epoch]*100:.1f}%"
                )
                if wandb_run_is_available():
                    wandb.log(
                        {
                            f"losses/loss__{sample_id}__ensemble_idx_{ensemble_idx}": tta_losses[
                                epoch
                            ]
                        },
                        step=global_idx,
                    )
                    wandb.log(
                        {
                            f"scores/eval_dice__{sample_id}__ensemble_idx_{ensemble_idx}": eval_dices[
                                epoch
                            ]
                        },
                        step=global_idx,
                    )

            tta_parameters = [model.state_dict()]
            torch.save(tta_parameters, tta_parameters_save_path)

            if not wandb_run_is_available():
                plot_run_results(
                    sub_dir_tta, sample_id, ensemble_idx, tta_losses, eval_dices
                )

            if debug:
                break
        # End of ensemble loop

    print("\n\n# Starting inference")
    all_prediction_save_paths = []

    for smp_idx in trange(num_samples, desc="Samples"):
        ensemble_parameter_paths = []

        tta_sample, tta_tens_list, param_sample_id, sub_dir_tta = get_sample_specs(
            smp_idx, inference_data, save_path, across_all_samples=False
        )
        tqdm.write(f"\nSample {param_sample_id}\n")
        tta_sample["data"] = get_imgs(tta_sample["data"].unsqueeze(0)).squeeze(0)

        # Update internal save path for nnUNet
        ofile = tta_sample["ofile"]
        new_ofile = str(save_path / ofile)
        tta_sample["ofile"] = new_ofile

        prediction_save_path = Path(new_ofile + ".nii.gz")
        prediction_save_path.parent.mkdir(exist_ok=True)

        for ensemble_idx in range(config["ensemble_count"]):
            ensemble_parameter_paths.append(
                get_parameters_save_path(sub_dir_tta, param_sample_id, ensemble_idx)
            )

        disable_internal_augmentation()
        model = get_model_from_network(network, modifier_fn_module)

        run_inference(config, [tta_sample], model, predictor, ensemble_parameter_paths)

        predicted_output_array, data_properties = sitk_io.read_seg(prediction_save_path)
        predicted_output = map_label(
            torch.as_tensor(predicted_output_array),
            get_map_idxs(label_mapping, optimized_labels, input_type="pretrain_labels"),
            input_format="argmaxed",
        ).squeeze(0)

        sitk_io.write_seg(
            predicted_output.numpy(), prediction_save_path, properties=data_properties
        )
        all_prediction_save_paths.append(prediction_save_path)

    # End of sample loop

    tqdm.write("\n\nEvaluating predictions")

    for pred_path in all_prediction_save_paths:
        pred_label_name = Path(pred_path).name
        if "outputTs" in Path(pred_path).parent.parts[-1]:
            path_mapped_target = save_path / "mapped_target_labelsTs" / pred_label_name
            path_orig_target = tta_data_dir / "labelsTs" / pred_label_name
        elif "outputTr" in Path(pred_path).parent.parts[-1]:
            path_mapped_target = save_path / "mapped_target_labelsTr" / pred_label_name
            path_orig_target = tta_data_dir / "labelsTr" / pred_label_name
        else:
            raise ValueError()

        if not path_orig_target.is_file():
            # No target available
            continue

        path_mapped_target.parent.mkdir(exist_ok=True)
        shutil.copy(path_orig_target, path_mapped_target)

        seg, sitk_stuff = sitk_io.read_seg(path_mapped_target)
        seg = torch.as_tensor(seg)
        mapped_seg = map_label(
            seg,
            get_map_idxs(label_mapping, optimized_labels, input_type="tta_labels"),
            input_format="argmaxed",
        ).squeeze(0)
        sitk_io.write_seg(mapped_seg.squeeze(0).numpy(), path_mapped_target, sitk_stuff)

    for bucket in ["Ts", "Tr"]:
        all_mapped_targets_path = save_path / f"mapped_target_labels{bucket}"
        all_pred_targets_path = save_path / f"tta_output{bucket}"

        if not all_mapped_targets_path.is_dir() or not all_pred_targets_path.is_dir():
            continue

        # Run postprocessing
        postprocess_results_fn = (
            modifier_fn_module.ModifierFunctions.postprocess_results_fn
        )
        postprocess_results_fn(all_pred_targets_path)

        summary_path = f"{save_path}/summary_{bucket}.json"
        compute_metrics_on_folder_simple(
            folder_ref=all_mapped_targets_path,
            folder_pred=all_pred_targets_path,
            labels=list(range(len(optimized_labels))),
            output_file=summary_path,
            num_processes=config["num_processes"],
            chill=True,
        )

        with open(summary_path, "r") as f:
            summary_json = json.load(f)
            final_mean_dice = summary_json["foreground_mean"]["Dice"]

        if wandb_run_is_available():
            wandb.log({f"scores/tta_dice_mean_{bucket}": final_mean_dice})


def calc_branch(
    branch_id,
    config,
    model,
    intensity_aug_func,
    identity_grid,
    patch_size,
    batch_size,
    label_mapping,
    optimized_labels,
    modifier_fn_module,
    imgs,
    device,
):
    assert branch_id in ["branch_a", "branch_b"]

    grad_context = (
        nullcontext if config["have_grad_in"] in ["branch_a", "both"] else torch.no_grad
    )

    modify_tta_output_after_mapping_fn = (
        modifier_fn_module.ModifierFunctions.modify_tta_output_after_mapping_fn
    )

    with grad_context():
        zero_grid = 0.0 * identity_grid

        imgs_aug = imgs

        if config["do_intensity_aug_in"] in [branch_id, "both"]:
            imgs_aug = intensity_aug_func(imgs_aug)
        else:
            imgs_aug = imgs_aug

        grid = zero_grid
        grid_inverse = zero_grid

        if config["spatial_aug_type"] == "affine" and config["do_spatial_aug_in"] in [
            branch_id,
            "both",
        ]:
            R, R_inverse = get_rand_affine(batch_size, flip=False)
            R, R_inverse = R.to(device), R_inverse.to(device)
            grid = grid + (
                F.affine_grid(R, [batch_size, 1] + patch_size, align_corners=False)
                - identity_grid
            )
            grid_inverse = grid_inverse + (
                F.affine_grid(
                    R_inverse, [batch_size, 1] + patch_size, align_corners=False
                )
                - identity_grid
            )

        if config["spatial_aug_type"] == "deformable" and config[
            "do_spatial_aug_in"
        ] in [branch_id, "both"]:
            grid_deformable, grid_deformable_inverse = get_disp_field(
                batch_size,
                patch_size,
                factor=0.5,
                interpolation_factor=5,
                device=device,
            )
            grid = grid + grid_deformable
            grid_inverse = grid_inverse + grid_deformable_inverse

        if config["do_spatial_aug_in"] in [branch_id, "both"]:
            grid = grid + identity_grid
            imgs_aug = F.grid_sample(
                imgs_aug, grid, padding_mode="border", align_corners=False
            )

        if branch_id == "branch_a":
            model.apply(buffer_running_stats)
        elif branch_id == "branch_b":
            model.apply(apply_running_stats)
        else:
            raise ValueError()

        branch_target = model(imgs_aug)
        branch_target = map_label(
            branch_target,
            get_map_idxs(label_mapping, optimized_labels, input_type="pretrain_labels"),
            input_format="logits",
        )
        branch_target = modify_tta_output_after_mapping_fn(branch_target)

        if isinstance(branch_target, tuple):
            branch_target = branch_target[0]

        if config["do_spatial_aug_in"] in [branch_id, "both"]:
            grid_inverse = grid_inverse + identity_grid
            branch_target = F.grid_sample(
                branch_target, grid_inverse, align_corners=False
            )
        else:
            branch_target = branch_target

        return branch_target



import json
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1.axes_grid import ImageGrid
import numpy as np
import torch

from nnunetv2.imageio.simpleitk_reader_writer import SimpleITKIO

from dg_tta.tta.config_log_utils import (
    get_data_filepaths,
    get_dgtta_colormap,
    get_resources_dir,
)
from dg_tta.utils import check_dga_root_is_set


def read_image(source_data_paths, path_idx):
    if source_data_paths is None:
        return None, None

    source_img, source_sitk_stuff = SimpleITKIO().read_images(
        source_data_paths[path_idx : path_idx + 1]
    )
    source_img = source_img[0]
    return torch.tensor(source_img)[None, None, :], source_sitk_stuff


def get_target_imgs_datapaths():
    check_dga_root_is_set()
    buckets = ["imagesTr", "imagesTs"]
    with open("tta_plan.json", "r") as f:
        tta_plan = json.load(f)
    target_dataset_name = tta_plan["__tta_dataset_name__"]
    target_data_paths = []

    for buc in buckets:
        target_data_paths.extend(get_data_filepaths(target_dataset_name, buc))
    return target_data_paths


def get_source_imgs_datapaths():
    check_dga_root_is_set()
    buckets = ["imagesTr", "imagesTs"]
    with open("tta_plan.json", "r") as f:
        tta_plan = json.load(f)
    source_dataset_name = tta_plan["__pretrained_dataset_name__"]

    if source_dataset_name.startswith("TS104"):
        return "TS104"

    source_data_paths = []
    for buc in buckets:
        source_data_paths.extend(get_data_filepaths(source_dataset_name, buc))
    return source_data_paths


def get_orient_imgs(img):
    def get_axes_idxs(axis_size):
        NUM_IDXS = 16
        return np.linspace(0, axis_size - 1, NUM_IDXS).round().astype(int)

    img = img.squeeze(0, 1)
    D, H, W = img.shape
    slices = dict(HW=[], DW=[], DH=[])
    for d in get_axes_idxs(D):
        slices["HW"].append(img[d, :, :])
    for h in get_axes_idxs(H):
        slices["DW"].append(img[:, h, :])
    for w in get_axes_idxs(W):
        slices["DH"].append(img[:, :, w])
    return slices


def clear_axis(ax):
    ax.get_yaxis().set_ticks([])
    ax.get_xaxis().set_ticks([])
    ax.grid(False)
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["left"].set_visible(False)


def get_spacing_ratio(sitk_stuff, axis_idx):
    rolled_spacing = np.roll(np.array(sitk_stuff["spacing"]), axis_idx)
    return rolled_spacing[1] / rolled_spacing[0]


def show_image_overview(img, sitk_stuff, fig_inch_size=5.0):
    orient_imgs = get_orient_imgs(img)
    vmin, vmax = img.min(), img.max()

    dpi = 150.0
    large_text_size = fig_inch_size * 10
    small_text_size = fig_inch_size * 2
    cmap = get_dgtta_colormap()

    fig, fig_axes = plt.subplots(2, 2, figsize=(fig_inch_size, fig_inch_size), dpi=dpi)
    fig.subplots_adjust(wspace=0.01, hspace=0.01)

    for fig_ax in fig_axes.flatten():
        clear_axis(fig_ax)

    for idx, (orientation, slices) in enumerate(orient_imgs.items()):
        current_fig_ax = fig_axes.flatten()[idx]
        grid_axes = ImageGrid(fig, 221 + idx, nrows_ncols=(4, 4), axes_pad=0.0)

        for ax, slc in zip(grid_axes, slices):
            ax.imshow(slc, cmap="gray", vmin=vmin, vmax=vmax)
            ax.set_aspect(get_spacing_ratio(sitk_stuff, idx))

        for ax in grid_axes:
            clear_axis(ax)

        plt.figtext(
            0.5,
            0.5,
            orientation,
            ha="center",
            va="center",
            fontsize=large_text_size,
            color=cmap(idx / 3),
            transform=current_fig_ax.transAxes,
        )

    for fig_ax in fig_axes.flatten():
        fig_ax.set_facecolor("black")

    fig.set_facecolor("black")
    plt.figtext(
        0.5,
        0.5,
        f"Spacing: {[np.round(sp,1) for sp in sitk_stuff['spacing']]}",
        color="white",
        ha="center",
        va="center",
        fontsize=small_text_size,
        transform=fig_axes.flatten()[3].transAxes,
    )
    # plt.savefig("out.png", bbox_inches="tight", pad_inches=0)
    plt.show()
    plt.close(fig)


def show_ts104_image():
    img_path = get_resources_dir() / "TS104_input_view.png"
    fig_inch_size = 7.0
    fig, ax = plt.subplots(dpi=150.0, figsize=(fig_inch_size, fig_inch_size))
    fig.set_facecolor("black")
    img = matplotlib.image.imread(img_path)
    ax.imshow(img)
    clear_axis(ax)
    ax.set_facecolor("black")
    plt.show()
    plt.close(fig)



import os
import sys
import pathlib
from pathlib import Path
import importlib
import subprocess
import shutil
import inspect
import json
from contextlib import contextmanager

import matplotlib
from matplotlib import pyplot as plt

if importlib.util.find_spec("wandb"):
    import wandb
import torch

from nnunetv2.paths import nnUNet_raw, nnUNet_results
from nnunetv2.utilities.dataset_name_id_conversion import maybe_convert_to_dataset_name

import dg_tta

TEMPLATE_PLAN = dict(
    tta_across_all_samples=False,
    tta_eval_patches=1,
    batch_size=1,
    patches_to_be_accumulated=16,
    lr=1e-5,
    ensemble_count=3,
    epochs=12,
    start_tta_at_epoch=1,
    intensity_aug_function="GIN",  # ['GIN', 'disabled']
    spatial_aug_type="affine",  # ['affine', 'deformable']
    params_with_grad="all",  # all, norms, encoder
    have_grad_in="branch_a",  # ['branch_a', 'branch_b', 'both']
    do_intensity_aug_in="none",  # ['branch_a', 'branch_b', 'both', 'none']
    do_spatial_aug_in="both",  # ['branch_a', 'branch_b', 'both', 'none']
    num_processes=1,
    wandb_mode="disabled",
)


class ModifierFunctions:
    def __init__(self):
        pass

    @staticmethod
    def modify_tta_input_fn(image: torch.Tensor):
        assert image.ndim == 5  # B,1,D,H,W
        # This function will be called on the input that is fed to the model
        return image

    @staticmethod
    def modfify_tta_model_output_fn(pred_label: torch.Tensor):
        assert pred_label.ndim == 5  # B,C,D,H,W
        # This function will be called directly after model prediction
        return pred_label

    @staticmethod
    def modify_tta_output_after_mapping_fn(mapped_label: torch.Tensor):
        assert mapped_label.ndim == 5  # B,MAPPED_C,D,H,W
        # This function will be called after model prediction when labels are mapped
        # to the target label numbers/ids.
        return mapped_label

    @staticmethod
    def postprocess_results_fn(results_dir: pathlib.Path):
        pass
        # This function will be called on the final output directory.


def wandb_run(wandb_project_name, tta_fn, **kwargs):
    config = kwargs["config"]
    with wandb.init(
        project=wandb_project_name,
        name=kwargs["run_name"],
        mode=config["wandb_mode"],
        config=config,
    ):
        kwargs["config"] = wandb.config
        tta_fn(**kwargs)
    wandb.finish()
    torch.cuda.empty_cache()


def get_tta_folders(
    pretrained_dataset_id,
    tta_dataset_id,
    pretrainer,
    pretrainer_config,
    pretrainer_fold,
):
    root_dir = Path(os.environ["DG_TTA_ROOT"])

    pretrainer, pretrainer_config, pretrainer_fold = check_dataset_pretrain_config(
        pretrained_dataset_id, pretrainer, pretrainer_config, pretrainer_fold
    )

    # Get dataset names
    tta_dataset_name = maybe_convert_to_dataset_name(tta_dataset_id)

    if isinstance(pretrained_dataset_id, int):
        pretrained_dataset_name = maybe_convert_to_dataset_name(pretrained_dataset_id)
    else:
        pretrained_dataset_name = pretrained_dataset_id

    fold_folder = (
        f"fold_{pretrainer_fold}" if pretrainer_fold != "all" else pretrainer_fold
    )
    map_folder = f"Pretrained_{pretrained_dataset_name}_at_{tta_dataset_name}"
    pretrainer_folder = f"{pretrainer}__{pretrainer_config}"

    plan_dir = root_dir / "plans" / map_folder / pretrainer_folder / fold_folder
    results_dir = root_dir / "results" / map_folder / pretrainer_folder / fold_folder

    tta_data_dir = Path(nnUNet_raw, tta_dataset_name)

    return (
        tta_data_dir,
        plan_dir,
        results_dir,
        pretrained_dataset_name,
        tta_dataset_name,
    )


def check_dataset_pretrain_config(
    pretrained_dataset_id, pretrainer, pretrainer_config, pretrainer_fold
):
    assert pretrained_dataset_id in [
        "TS104_GIN",
        "TS104_MIND",
        "TS104_GIN_MIND",
    ] or isinstance(pretrained_dataset_id, int)

    if isinstance(pretrained_dataset_id, int):
        # Check fold specifier
        assert pretrainer is not None
        assert pretrainer_config is not None
        assert pretrainer_fold == "all" or isinstance(pretrainer_fold, int)
    else:
        if pretrained_dataset_id == "TS104_GIN":
            pretrainer = "nnUNetTrainer_GIN"
            pretrainer_config = "3d_fullres"
            pretrainer_fold = "0"

        elif pretrained_dataset_id == "TS104_MIND":
            pretrainer = "nnUNetTrainer_MIND"
            pretrainer_config = "3d_fullres"
            pretrainer_fold = "0"

        elif pretrained_dataset_id == "TS104_GIN_MIND":
            pretrainer = "nnUNetTrainer_GIN_MIND"
            pretrainer_config = "3d_fullres"
            pretrainer_fold = "0"

        else:
            raise ValueError()

    return pretrainer, pretrainer_config, pretrainer_fold


def prepare_tta(
    pretrained_dataset_id,
    tta_dataset_id,
    pretrainer=None,
    pretrainer_config=None,
    pretrainer_fold=None,
    tta_dataset_bucket="imagesTs",
):
    root_dir = Path(os.environ["DG_TTA_ROOT"])
    assert root_dir.is_dir()

    # Create directories
    (
        _,
        plan_dir,
        results_dir,
        pretrained_dataset_name,
        tta_dataset_name,
    ) = get_tta_folders(
        pretrained_dataset_id,
        tta_dataset_id,
        pretrainer,
        pretrainer_config,
        pretrainer_fold,
    )

    shutil.rmtree(plan_dir, ignore_errors=True)
    plan_dir.mkdir(exist_ok=True, parents=True)
    results_dir.mkdir(exist_ok=True, parents=True)

    if isinstance(pretrained_dataset_id, str):
        target_path, weights_file_path = download_pretrained_weights(
            pretrained_dataset_id
        )
        with open(target_path / "dataset.json", "r") as f:
            pretrained_classes = json.load(f)["labels"]

    else:
        # Get label mappings from nnUNet task
        raw_pretrained_dataset_dir = Path(nnUNet_raw, pretrained_dataset_name)
        with open(raw_pretrained_dataset_dir / "dataset.json", "r") as f:
            pretrained_classes = json.load(f)["labels"]

        # Get weights file
        fold_dir = (
            f"fold_{pretrainer_fold}" if pretrainer_fold != "all" else pretrainer_fold
        )
        results_pretrained_dataset_dir = Path(
            nnUNet_results,
            pretrained_dataset_name,
            f"{pretrainer}__nnUNetPlans__{pretrainer_config}",
            fold_dir,
        )
        weights_file_path = results_pretrained_dataset_dir / "checkpoint_final.pth"

        if not weights_file_path.is_file():
            raise FileNotFoundError(
                f"Could not find weights file at {weights_file_path}"
            )

    raw_tta_dataset_dir = Path(nnUNet_raw, tta_dataset_name)

    # Load tta_task classes
    with open(raw_tta_dataset_dir / "dataset.json", "r") as f:
        tta_dataset_classes = json.load(f)["labels"]

    # Dump label mappings
    with open(plan_dir / f"{pretrained_dataset_name}_label_mapping.json", "w") as f:
        json.dump(pretrained_classes, f, indent=4)
    with open(plan_dir / f"{tta_dataset_name}_label_mapping.json", "w") as f:
        json.dump(tta_dataset_classes, f, indent=4)

    # Create plan
    initial_plan = TEMPLATE_PLAN.copy()
    initial_plan["__pretrained_dataset_name__"] = pretrained_dataset_name
    initial_plan["__tta_dataset_name__"] = tta_dataset_name
    initial_plan["pretrained_weights_filepath"] = str(weights_file_path)

    # Retrive possible labels to be optimized (may require manual tweaking later)
    intersection_classes = list(
        set(pretrained_classes.keys()).intersection(set(tta_dataset_classes))
    )
    assert (
        "background" in intersection_classes
    ), "Background class must be present in both datasets!"
    intersection_classes.sort()
    intersection_classes.remove("background")
    intersection_classes.insert(0, "background")
    initial_plan["optimized_labels"] = intersection_classes

    # Retrive filepath of tta_task data
    tta_data_filepaths = get_data_filepaths(tta_dataset_name, tta_dataset_bucket)
    initial_plan["tta_data_filepaths"] = [str(fp) for fp in tta_data_filepaths]

    # Dump plan
    with open(plan_dir / "tta_plan.json", "w") as f:
        json.dump(initial_plan, f, indent=4)

    # Dump modifier functions
    modifier_src = inspect.getsource(ModifierFunctions)

    with open(plan_dir / "modifier_functions.py", "w") as f:
        f.write("import pathlib\n")
        f.write("import torch\n\n")
        f.write(modifier_src)

    copy_check_tta_input_notebook(plan_dir)

    print(
        f"\nPreparation done. You can edit the plan, modifier functions and optimized labels in {plan_dir} prior to running TTA."
    )


def get_resources_dir():
    return Path(dg_tta.__file__).parent / "__resources__"


def download_pretrained_weights(pretrained_dataset_id):
    pretrained_weights_dir = Path(os.environ["DG_TTA_ROOT"]) / "_pretrained_weights"

    if pretrained_dataset_id == "TS104_GIN":
        pretrainer_dir = "nnUNetTrainer_GIN__nnUNetPlans__3d_fullres"
        dl_link = "https://cloud.imi.uni-luebeck.de/s/ERK6Wic3D95qDKz/download"

    elif pretrained_dataset_id == "TS104_MIND":
        pretrainer_dir = "nnUNetTrainer_MIND__nnUNetPlans__3d_fullres"
        dl_link = "https://cloud.imi.uni-luebeck.de/s/LZByo9m3A5c6Dki/download"

    elif pretrained_dataset_id == "TS104_GIN_MIND":
        pretrainer_dir = "nnUNetTrainer_GIN_MIND__nnUNetPlans__3d_fullres"
        dl_link = "https://cloud.imi.uni-luebeck.de/s/dkGdfFGwbnzWya4/download"

    else:
        raise ValueError()

    dummy_results_path = get_resources_dir() / "dummy_results" / pretrainer_dir
    target_path = pretrained_weights_dir / pretrainer_dir
    target_path_weights = target_path / "fold_0" / "checkpoint_final.pth"

    target_path.mkdir(exist_ok=True)
    target_path_weights.parent.mkdir(exist_ok=True)

    # Copy dummy pretraining results (folder structure and nnUNet fils)
    shutil.copytree(dummy_results_path, target_path, dirs_exist_ok=True)

    if not target_path_weights.exists():
        subprocess.run(["wget", dl_link, "-O", target_path_weights])

    return target_path, target_path_weights


def get_global_idx(list_of_tuple_idx_max):
    # Get global index e.g. 2250 for ensemble_idx=2, epoch_idx=250 @ max_epochs<1000
    global_idx = 0
    next_multiplier = 1

    # Smallest identifier tuple last!
    for idx, max_of_idx in reversed(list_of_tuple_idx_max):
        global_idx = global_idx + next_multiplier * idx
        next_multiplier = next_multiplier * 10 ** len(str(int(max_of_idx)))
    return global_idx


def load_current_modifier_functions(plan_dir):
    mod_path = Path(plan_dir / "modifier_functions.py")
    spec = importlib.util.spec_from_file_location(
        "dg_tta.current_modifier_functions", mod_path
    )
    dyn_mod = importlib.util.module_from_spec(spec)
    sys.modules["dg_tta.current_modifier_functions"] = dyn_mod
    spec.loader.exec_module(dyn_mod)

    return dyn_mod


def get_data_filepaths(tta_dataset_name, tta_dataset_bucket):
    raw_tta_dataset_dir = Path(nnUNet_raw, tta_dataset_name)
    if tta_dataset_bucket == "imagesTr":
        source_folders = [raw_tta_dataset_dir / "imagesTr"]
    elif tta_dataset_bucket == "imagesTs":
        source_folders = [raw_tta_dataset_dir / "imagesTs"]
    elif tta_dataset_bucket == "imagesTrAndTs":
        source_folders = [
            raw_tta_dataset_dir / "imagesTr",
            raw_tta_dataset_dir / "imagesTs",
        ]

    file_list = []
    for src_fld in source_folders:
        file_list.extend(filter(lambda x: x.is_file(), src_fld.iterdir()))

    return file_list


def wandb_run_is_available():
    return (
        importlib.util.find_spec("wandb") is not None
        and wandb.run is not None
        and not wandb.run.disabled
    )


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def get_dgtta_colormap():
    hi_1 = "#248888"
    hi_2 = "#79DCF0"
    hi_3 = "#e7475e"
    hi_4 = "#f0d879"
    return matplotlib.colors.LinearSegmentedColormap.from_list(
        "", [hi_3, hi_4, hi_2, hi_1]
    )


def plot_run_results(save_path, sample_id, ensemble_idx, tta_losses, eval_dices):
    fig, ax_one = plt.subplots()
    ax_two = ax_one.twinx()
    cmap = get_dgtta_colormap()
    c1, c2 = cmap(0.0), cmap(0.8)
    ax_one.plot(tta_losses, label="loss", c=c1)
    ax_one.set_yticks([tta_losses.min(), tta_losses.max()])
    ax_one.set_xlim(0, len(tta_losses) - 1)
    ax_one.set_ylabel("Soft-Dice Loss", c=c1)
    ax_one.tick_params(axis="y", colors=c1)
    ax_one.set_xlabel("TTA Epoch")
    ax_one.grid(axis="y", linestyle="--", linewidth=0.5)
    ax_one.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter("%.3f"))

    ax_two.plot(eval_dices * 100, label="eval_dices", c=c2)
    ax_two.set_yticks([eval_dices.min() * 100, eval_dices.max() * 100])
    ax_two.set_ylabel("Pseudo Dice in %", c=c2)
    ax_two.tick_params(axis="y", colors=c2)
    ax_two.yaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter("%.1f"))
    fig.suptitle(f"{sample_id} (ensemble_idx={ensemble_idx})")
    split_sample_id = sample_id.split('/')[-1]
    tta_plot_save_path = (
        save_path / f"{split_sample_id}__ensemble_idx_{ensemble_idx}_tta_results.png"
    )
    fig.savefig(tta_plot_save_path)
    fig.tight_layout()
    plt.close(fig)


def copy_check_tta_input_notebook(plan_dir):
    NB_FILENAME = "check_tta_input.ipynb"
    shutil.copyfile(
        get_resources_dir() / NB_FILENAME,
        plan_dir / NB_FILENAME,
    )


def get_parameters_save_path(save_path, sample_id, ensemble_idx):
    sample_id = sample_id.split('/')[-1]
    tta_parameters_save_path = (
        save_path / f"{sample_id}__ensemble_idx_{ensemble_idx}_tta_parameters.pt"
    )
    return tta_parameters_save_path


from itertools import chain
from typing import List, Union
from pathlib import Path
from copy import deepcopy

import torch
import numpy as np

from nnunetv2.inference.predict_from_raw_data import nnUNetPredictor
from nnunetv2.utilities.plans_handling.plans_handler import (
    PlansManager,
    ConfigurationManager,
)
from nnunetv2.utilities.label_handling.label_handling import convert_labelmap_to_one_hot

from dg_tta.tta.config_log_utils import suppress_stdout


def get_data_iterator(
    config, predictor, tta_data_filepaths, dataset_raw_path, tta_dataset_bucket
):
    assert tta_dataset_bucket in ["imagesTs", "imagesTr"]

    list_of_lists = [
        [_path]
        for _path in tta_data_filepaths
        if Path(_path).parts[-2] == tta_dataset_bucket
    ]

    label_folder = "labelsTs" if tta_dataset_bucket == "imagesTs" else "labelsTr"
    output_folder = (
        "tta_outputTs" if tta_dataset_bucket == "imagesTs" else "tta_outputTr"
    )

    (
        list_of_lists_or_source_folder,
        output_filename_truncated,
        seg_from_prev_stage_files,
    ) = predictor._manage_input_and_output_lists(
        list_of_lists, output_folder, dataset_raw_path / label_folder
    )

    if len(list_of_lists_or_source_folder) == 0:
        return iter(()), 0

    seg_from_prev_stage_files = [
        s if Path(s).is_file() else None for s in seg_from_prev_stage_files
    ]
    data_iterator = get_data_iterator_from_lists_of_filenames(
        predictor,
        list_of_lists_or_source_folder,
        seg_from_prev_stage_files,
        output_filename_truncated,
    )
    return data_iterator, len(list_of_lists)


def load_tta_data(config, dataset_raw_path, predictor, tta_across_all_samples=False):
    with suppress_stdout():
        ts_iterator, ts_data_len = get_data_iterator(
            config,
            predictor,
            config["tta_data_filepaths"],
            dataset_raw_path,
            "imagesTs",
        )
        tr_iterator, tr_data_len = get_data_iterator(
            config,
            predictor,
            config["tta_data_filepaths"],
            dataset_raw_path,
            "imagesTr",
        )

    if tta_across_all_samples:
        data = list(ts_iterator) + list(tr_iterator), ts_data_len + tr_data_len
    else:
        data = chain(ts_iterator, tr_iterator), ts_data_len + tr_data_len

    return data


def load_network(weights_file, device):
    pretrained_weights_filepath = Path(*Path(weights_file).parts[:-2])
    fold = Path(weights_file).parts[-2].replace("fold_", "")
    use_folds = [int(fold)] if fold.isnumeric() else fold
    checkpoint_name = Path(weights_file).parts[-1]
    configuration = Path(weights_file).parts[-3].split("__")[-1]

    perform_everything_on_gpu = True
    verbose = False

    predictor = nnUNetPredictor(
        perform_everything_on_gpu=perform_everything_on_gpu,
        device=device,
        verbose_preprocessing=verbose,
    )

    predictor.initialize_from_trained_model_folder(
        pretrained_weights_filepath, use_folds, checkpoint_name
    )

    parameters = predictor.list_of_parameters
    plans_manager = predictor.plans_manager
    network = predictor.network
    patch_size = plans_manager.get_configuration(configuration).patch_size

    return predictor, patch_size, network, parameters


def run_inference(config, tta_data, model, predictor, all_tta_parameter_paths):
    save_probabilities = False
    num_processes_segmentation_export = config["num_processes"]

    tta_parameters = []
    for _path in all_tta_parameter_paths:
        tta_parameters.extend(torch.load(_path))

    predictor.network = deepcopy(model)
    predictor.list_of_parameters = tta_parameters
    predictor.predict_from_data_iterator(
        tta_data, save_probabilities, num_processes_segmentation_export
    )


def get_data_iterator_from_lists_of_filenames(
    predictor,
    input_list_of_lists: List[List[str]],
    seg_from_prev_stage_files: Union[List[str], None],
    output_filenames_truncated: Union[List[str], None],
):
    # Adapted from nnunetv2/inference/predict_from_raw_data.py
    return preprocessing_iterator_fromfiles(
        input_list_of_lists,
        seg_from_prev_stage_files,
        output_filenames_truncated,
        predictor.plans_manager,
        predictor.dataset_json,
        predictor.configuration_manager,
        predictor.verbose_preprocessing,
    )


def preprocessing_iterator_fromfiles(
    list_of_lists: List[List[str]],
    list_of_segs_from_prev_stage_files: Union[None, List[str]],
    output_filenames_truncated: Union[None, List[str]],
    plans_manager: PlansManager,
    dataset_json: dict,
    configuration_manager: ConfigurationManager,
    verbose: bool = False,
):
    # Adapted from nnunetv2/inference/data_iterators.py
    num_samples = len(list_of_lists)
    for smp_idx in range(num_samples):
        item = preprocess_fromfile(
            list_of_lists[smp_idx],
            list_of_segs_from_prev_stage_files[smp_idx],
            output_filenames_truncated[smp_idx],
            plans_manager,
            dataset_json,
            configuration_manager,
            verbose,
        )
        yield item


def preprocess_fromfile(
    list_of_im_file: List[str],
    seg_from_prev_stage_file: str,
    output_filename_truncated: str,
    plans_manager: PlansManager,
    dataset_json: dict,
    configuration_manager: ConfigurationManager,
    verbose: bool = False,
):
    # Adapted from nnunetv2/inference/data_iterators.py
    label_manager = plans_manager.get_label_manager(dataset_json)
    preprocessor = configuration_manager.preprocessor_class(verbose=verbose)

    data, seg, data_properties = preprocessor.run_case(
        list_of_im_file,
        seg_from_prev_stage_file,
        plans_manager,
        configuration_manager,
        dataset_json,
    )

    if seg_from_prev_stage_file is not None:
        seg_onehot = convert_labelmap_to_one_hot(
            seg[0], label_manager.foreground_labels, data.dtype
        )
        data = np.vstack((data, seg_onehot))

    data = torch.from_numpy(data).contiguous().float()

    item = {
        "data": data,
        "data_properties": data_properties,
        "ofile": output_filename_truncated,
    }
    return item



import os
from pathlib import Path


def enable_internal_augmentation():
    os.environ["DG_TTA_INTERNAL_AUGMENTATION"] = "true"


def disable_internal_augmentation():
    os.environ["DG_TTA_INTERNAL_AUGMENTATION"] = "false"


def check_internal_augmentation_disabled():
    assert os.environ.get("DG_TTA_INTERNAL_AUGMENTATION").lower() != "true"


def get_internal_augmentation_enabled():
    return os.environ.get("DG_TTA_INTERNAL_AUGMENTATION").lower() == "true"


def check_dga_root_is_set(soft_check=False):
    prompt = "Please define an existing root directory for DG-TTA by setting DG_TTA_ROOT."
    check = Path(
        os.environ.get("DG_TTA_ROOT", "_")
    ).is_dir()

    if soft_check and not check:
        print(prompt)
        return

    assert check, prompt



import torch
import torch.nn.functional as F

from dg_tta.mind import MIND3D
from dg_tta.gin import gin_aug


def get_rf_field(
    num_batch, size_3d, interpolation_factor=4, num_fields=4, device="cpu"
):
    rf_field = F.interpolate(
        F.avg_pool3d(
            F.avg_pool3d(
                F.avg_pool3d(
                    torch.randn(
                        num_batch,
                        num_fields,
                        size_3d[0] // interpolation_factor,
                        size_3d[1] // interpolation_factor,
                        size_3d[2] // interpolation_factor,
                        device=device,
                    ),
                    interpolation_factor,
                    stride=1,
                    padding=interpolation_factor // 2,
                ),
                interpolation_factor,
                stride=1,
                padding=interpolation_factor // 2,
            ),
            interpolation_factor,
            stride=1,
            padding=interpolation_factor // 2,
        ),
        size=size_3d,
        mode="trilinear",
    )
    rf_field -= rf_field.mean((-3, -2, -1), keepdim=True)
    rf_field /= 1e-3 + rf_field.view(num_batch * num_fields, -1).std(1).view(
        num_batch, num_fields, 1, 1, 1
    )

    return rf_field


def calc_consistent_diffeomorphic_field(
    disp_field,
    inverse_disp_field,
    time_steps=1,
    ensure_inverse_consistency=True,
    iter_steps_override=None,
):
    # https://github.com/multimodallearning/convexAdam/blob/76a595914eb21ea17795e6cd19503ab447f0ea6b/l2r_2021_convexAdam_task1_docker.py#L166
    # https://github.com/cwmok/LapIRN/blob/d8f96770a704b1f190955cc26297c7b01a270b0a/Code/miccai2020_model_stage.py#L761

    # Vincent Arsigny, Olivier Commowick, Xavier Pennec, Nicholas Ayache: A Log-Euclidean Framework for Statistics on Diffeomorphisms
    B, C, D, H, W = disp_field.size()
    dimension_correction = torch.tensor([D, H, W], device=disp_field.device).view(
        1, 3, 1, 1, 1
    )
    dt = 1.0 / time_steps

    with torch.no_grad():
        identity = (
            F.affine_grid(
                torch.eye(3, 4).unsqueeze(0), (1, 1, D, H, W), align_corners=True
            )
            .permute(0, 4, 1, 2, 3)
            .to(disp_field)
        )
        if ensure_inverse_consistency:
            out_disp_field = (
                disp_field / dimension_correction / (2**time_steps) * dt
            ).clone()
            out_inverse_disp_field = (
                inverse_disp_field / dimension_correction / (2**time_steps) * dt
            ).clone()

            for _ in range(
                time_steps if not iter_steps_override else iter_steps_override
            ):
                ds = out_disp_field.clone()
                inverse_ds = out_inverse_disp_field.clone()
                out_disp_field = +0.5 * ds - 0.5 * F.grid_sample(
                    inverse_ds,
                    (identity + ds).permute(0, 2, 3, 4, 1),
                    padding_mode="border",
                    align_corners=True,
                )

                out_inverse_disp_field = +0.5 * inverse_ds - 0.5 * F.grid_sample(
                    ds,
                    (identity + inverse_ds).permute(0, 2, 3, 4, 1),
                    padding_mode="border",
                    align_corners=True,
                )
            out_disp_field = out_disp_field * 2**time_steps * dimension_correction
            out_inverse_disp_field = (
                out_inverse_disp_field * 2**time_steps * dimension_correction
            )

        else:
            # https://github.com/cwmok/LapIRN/blob/d8f96770a704b1f190955cc26297c7b01a270b0a/Code/miccai2020_model_stage.py#L761

            ds_dt = (
                disp_field / dimension_correction / (2**time_steps)
            )  # velocity = ds/dt
            inverse_ds_dt = (
                inverse_disp_field / dimension_correction / (2**time_steps)
            )
            ds = ds_dt * dt
            inverse_ds = inverse_ds_dt * dt

            for _ in range(
                time_steps if not iter_steps_override else iter_steps_override
            ):
                ds = ds + F.grid_sample(
                    ds,
                    (identity + ds).permute(0, 2, 3, 4, 1),
                    mode="bilinear",
                    padding_mode="zeros",
                    align_corners=True,
                )
                inverse_ds = inverse_ds + F.grid_sample(
                    inverse_ds,
                    (identity + inverse_ds).permute(0, 2, 3, 4, 1),
                    mode="bilinear",
                    padding_mode="zeros",
                    align_corners=True,
                )

            out_disp_field = ds * dimension_correction
            out_inverse_disp_field = inverse_ds * dimension_correction

    return out_disp_field, out_inverse_disp_field


def get_disp_field(
    batch_num, size_3d, factor=0.1, interpolation_factor=5, device="cpu"
):
    field = get_rf_field(
        batch_num,
        size_3d,
        alternating_fields=False,
        num_fields=3,
        interpolation_factor=interpolation_factor,
        device=device,
    )
    STEPS = 5
    disp_field, inverse_disp_field = calc_consistent_diffeomorphic_field(
        field * factor, torch.zeros_like(field), STEPS, ensure_inverse_consistency=True
    )
    return disp_field.permute(0, 2, 3, 4, 1), inverse_disp_field.permute(0, 2, 3, 4, 1)


def get_rand_affine(batch_size, strength=0.05, flip=False):
    affine = torch.cat(
        (
            torch.randn(batch_size, 3, 4) * strength + torch.eye(3, 4).unsqueeze(0),
            torch.tensor([0, 0, 0, 1]).view(1, 1, 4).repeat(batch_size, 1, 1),
        ),
        1,
    )

    if flip:
        flip_affine = torch.diag(
            torch.cat([(2 * (torch.rand(3) > 0.5).float() - 1), torch.tensor([1.0])])
        )
        affine = affine @ flip_affine
    return affine[:, :3], affine.inverse()[:, :3]


def gin_mind_aug(input):
    return MIND3D()(gin_aug(input))



from copy import deepcopy

from torch._dynamo import OptimizedModule

from dg_tta.tta.torch_utils import (
    register_forward_pre_hook_at_beginning,
    register_forward_hook_at_beginning,
    hookify,
)


def get_model_from_network(network, modifier_fn_module, parameters=None):
    model = deepcopy(network)

    if parameters is not None:
        if not isinstance(model, OptimizedModule):
            model.load_state_dict(parameters[0])
        else:
            model._orig_mod.load_state_dict(parameters[0])

    # Register hook that modifies the input prior to custom augmentation
    modify_tta_input_fn = modifier_fn_module.ModifierFunctions.modify_tta_input_fn
    register_forward_pre_hook_at_beginning(
        model, hookify(modify_tta_input_fn, "forward_pre_hook")
    )

    # Register hook that modifies the output of the model
    modfify_tta_model_output_fn = (
        modifier_fn_module.ModifierFunctions.modfify_tta_model_output_fn
    )
    register_forward_hook_at_beginning(
        model, hookify(modfify_tta_model_output_fn, "forward_hook")
    )

    return model


running_stats_buffer = {}


def buffer_running_stats(m):
    _id = id(m)
    if (
        hasattr(m, "running_mean")
        and hasattr(m, "running_var")
        and not _id in running_stats_buffer
    ):
        if m.running_mean is not None and m.running_var is not None:
            running_stats_buffer[_id] = [m.running_mean.data, m.running_var.data]


def apply_running_stats(m):
    _id = id(m)
    if (
        hasattr(m, "running_mean")
        and hasattr(m, "running_var")
        and _id in running_stats_buffer
    ):
        m.running_mean.data.copy_(other=running_stats_buffer[_id][0])
        m.running_var.data.copy_(
            other=running_stats_buffer[_id][1]
        )  # Copy into .data to prevent backprop errors
        del running_stats_buffer[_id]



import torch
import torch.nn.functional as F


def filter1D(img, weight, dim, padding_mode="replicate"):
    B, C, D, H, W = img.shape
    N = weight.shape[0]

    padding = torch.zeros(
        6,
    )
    padding[[4 - 2 * dim, 5 - 2 * dim]] = N // 2
    padding = padding.long().tolist()

    view = torch.ones(
        5,
    )
    view[dim + 2] = -1
    view = view.long().tolist()

    return F.conv3d(
        F.pad(img.view(B * C, 1, D, H, W), padding, mode=padding_mode),
        weight.view(view),
    ).view(B, C, D, H, W)


def smooth(img, sigma):
    device = img.device

    sigma = torch.tensor([sigma], device=device)
    N = torch.ceil(sigma * 3.0 / 2.0).long().item() * 2 + 1

    weight = torch.exp(
        -torch.pow(torch.linspace(-(N // 2), N // 2, N, device=device), 2)
        / (2 * torch.pow(sigma, 2))
    )
    weight /= weight.sum()

    img = filter1D(img, weight, 0)
    img = filter1D(img, weight, 1)
    img = filter1D(img, weight, 2)

    return img


def pdist(x, p=2):
    if p == 1:
        dist = torch.abs(x.unsqueeze(2) - x.unsqueeze(1)).sum(dim=3)
    elif p == 2:
        xx = (x**2).sum(dim=2).unsqueeze(2)
        yy = xx.permute(0, 2, 1)
        dist = xx + yy - 2.0 * torch.bmm(x, x.permute(0, 2, 1))
        dist[:, torch.arange(dist.shape[1]), torch.arange(dist.shape[2])] = 0
    return dist


class MIND2D_64:  # layout should be of size 2x64x1x2
    def __init__(self, image, layout, grid) -> None:
        # batch and channels should be equal to 1
        B, C, H, W = image.size()

        # smaller fixed length offsets for 64 MIND-SSC like features
        brief_layout3 = layout[0:1, 0:, :, :] * 0.15
        brief_layout4 = layout[1:2, 0:, :, :] * 0.15
        brief_layout4[:, :32, :, :] = 0
        fixed_length = 0.05
        brief_length = torch.sqrt(
            torch.sum((brief_layout3 - brief_layout4) ** 2, 3, keepdim=True)
        )
        brief_layout3 /= brief_length / fixed_length
        brief_layout4 /= brief_length / fixed_length

        img_patch = F.unfold(image, 5, padding=2).view(1, 25, H, W)
        brief_patch = torch.sum(
            (
                F.grid_sample(
                    img_patch,
                    brief_layout3 + grid.view(1, 1, -1, 2),
                    align_corners=True,
                )
                - F.grid_sample(
                    img_patch,
                    brief_layout4 + grid.view(1, 1, -1, 2),
                    align_corners=True,
                )
            )
            ** 2,
            1,
        )
        brief_patch -= brief_patch.min(1)[0]
        brief_patch /= torch.clamp_min(brief_patch.std(1), 1e-5)
        brief_patch = torch.exp(-brief_patch).view(1, -1, grid.size(1), grid.size(2))

        return brief_patch


class MIND3D(torch.nn.Module):
    def __init__(self, delta=1, sigma=1, randn_weighting=0.05) -> None:
        super().__init__()
        self.delta = delta
        self.sigma = sigma
        self.out_channels = 12
        # define start and end locations for self-similarity pattern
        six_neighbourhood = torch.tensor(
            [[0, 1, 1], [1, 1, 0], [1, 0, 1], [1, 1, 2], [2, 1, 1], [1, 2, 1]],
            dtype=torch.float,
        )

        # squared distances
        dist = pdist(six_neighbourhood.unsqueeze(0)).squeeze(0)

        # define comparison mask
        x, y = torch.meshgrid(torch.arange(6), torch.arange(6), indexing="ij")
        mask = (x > y).view(-1) & (dist == 2).view(-1)

        # build kernel
        idx_shift1 = (
            six_neighbourhood.unsqueeze(1).repeat(1, 6, 1).view(-1, 3)[mask, :].long()
        )
        idx_shift2 = (
            six_neighbourhood.unsqueeze(0).repeat(6, 1, 1).view(-1, 3)[mask, :].long()
        )
        mshift1 = torch.zeros((12, 1, 3, 3, 3))
        mshift1.view(-1)[
            torch.arange(12) * 27
            + idx_shift1[:, 0] * 9
            + idx_shift1[:, 1] * 3
            + idx_shift1[:, 2]
        ] = 1
        mshift2 = torch.zeros((12, 1, 3, 3, 3))
        mshift2.view(-1)[
            torch.arange(12) * 27
            + idx_shift2[:, 0] * 9
            + idx_shift2[:, 1] * 3
            + idx_shift2[:, 2]
        ] = 1
        self.rpad = torch.nn.ReplicationPad3d(delta)
        self.mshift1 = mshift1
        self.mshift2 = mshift2
        self.randn_weighting = randn_weighting

    def forward(self, img):
        # compute patch-ssd
        device = img.device

        edge_selection = F.conv3d(
            self.rpad(img), self.mshift1.to(device), dilation=self.delta
        ) - F.conv3d(self.rpad(img), self.mshift2.to(device), dilation=self.delta)

        edge_selection = edge_selection + self.randn_weighting * torch.randn_like(
            edge_selection
        )
        ssd = smooth(edge_selection**2, self.sigma)

        # MIND equation
        mind = ssd - torch.min(ssd, 1, keepdim=True)[0]
        mind_var = torch.mean(mind, 1, keepdim=True)
        mind_var = torch.clamp(
            mind_var, mind_var.mean() * 0.001, mind_var.mean() * 1000
        )
        mind /= mind_var
        mind = torch.exp(-mind)

        return mind


def mind_hook(module, input):
    return MIND3D().forward(*input)



import re
from pathlib import Path
import nnunetv2
from dg_tta import pretraining


def inject_dg_trainers_into_nnunet(num_epochs=1000):
    dg_trainer_paths = Path(pretraining.__file__).parent.glob("nnUNetTrainer*.py")
    target_dir = Path(nnunetv2.__path__[0], "training/nnUNetTrainer/variants/dg_tta/")
    target_dir.mkdir(exist_ok=True)
    (target_dir / "__init__.py").touch()  # Make directory a mdoule

    for tr in dg_trainer_paths:
        # open file
        with open(tr, "r") as f:
            tr_code = f.read()
        tr_code_with_set_epochs = re.sub(
            r"self\.num_epochs = \d+", f"self.num_epochs = {num_epochs}", tr_code
        )

        with open(target_dir / tr.name, "w") as f:
            f.write(tr_code_with_set_epochs)


def check_trainers_injected():
    target_dir = Path(nnunetv2.__path__[0], "training/nnUNetTrainer/variants/dg_tta/")
    assert (
        target_dir.exists()
    ), "DG trainers not injected into nnUNet module. Please inject trainers first."



import sys
import re
from pathlib import Path
import json
import argparse
import json
from datetime import datetime

import torch
import torch.nn.functional as F

import randomname

from nnunetv2.run.run_training import run_training_entry as nnunet_run_training_main

import dg_tta
from dg_tta.__build__ import inject_dg_trainers_into_nnunet, check_trainers_injected
from dg_tta.utils import check_dga_root_is_set
from dg_tta.tta.torch_utils import generate_label_mapping

from dg_tta.tta.config_log_utils import (
    wandb_run,
    load_current_modifier_functions,
    get_tta_folders,
    wandb_run_is_available,
)
from dg_tta.tta.tta import tta_main

PROJECT_NAME = "nnunet_tta"


class DGTTAProgram:
    def __init__(self):
        parser = argparse.ArgumentParser(
            description="DG-TTA for nnUNetv2",
            usage="""dgtta <command> [<args>]

        Commands are:
        inject_trainers Inject DG trainers into nnUNet module
        pretrain        Pretrain on a dataset with DG trainers
        prepare_tta     Prepare test-time adaptation
        run_tta         Run test-time adaptation
        """,
        )
        parser.add_argument("command", help="Subcommand to run")
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print("Unrecognized command")
            parser.print_help()
            exit(1)
        getattr(self, args.command)()

    def inject_trainers(self):
        parser = argparse.ArgumentParser(
            description="Inject DG-TTA trainers into nnUNet module code"
        )
        parser.add_argument(
            "--num_epochs", type=int, default=1000, help="Number of epochs to train"
        )
        args = parser.parse_args(sys.argv[2:])
        inject_dg_trainers_into_nnunet(args.num_epochs)

    def pretrain(self):
        check_trainers_injected()
        print("Dispatching into nnUNetv2_train.")
        sys.argv = sys.argv[2:]
        sys.argv.insert(0, "nnUNetv2_train")
        nnunet_run_training_main()

    def prepare_tta(self):
        check_trainers_injected()
        parser = argparse.ArgumentParser(
            description="Prepare DG-TTA", usage="""dgtta prepare_tta [-h]"""
        )

        parser.add_argument(
            "pretrained_dataset_id",
            help="""
                            Task ID for pretrained model.
                            Can be numeric or one of ['TS104_GIN', 'TS104_MIND', 'TS104_GIN_MIND']""",
        )
        parser.add_argument("tta_dataset_id", help="Task ID for TTA")
        parser.add_argument(
            "--pretrainer",
            help="Trainer to use for pretraining",
            default=None,
        )
        parser.add_argument(
            "--pretrainer_config",
            help="Fold ID of nnUNet model to use for pretraining",
            default="3d_fullres",
        )
        parser.add_argument(
            "--pretrainer_fold",
            help="Fold ID of nnUNet model to use for pretraining",
            default="0",
        )
        parser.add_argument(
            "--tta_dataset_bucket",
            help="""Can be one of ['imagesTr', 'imagesTs', 'imagesTrAndTs']""",
            default="imagesTs",
        )

        args = parser.parse_args(sys.argv[2:])
        pretrained_dataset_id = (
            int(args.pretrained_dataset_id)
            if args.pretrained_dataset_id.isnumeric()
            else args.pretrained_dataset_id
        )
        pretrainer_fold = (
            int(args.pretrainer_fold)
            if args.pretrainer_fold.isnumeric()
            else args.pretrainer_fold
        )

        dg_tta.tta.config_log_utils.prepare_tta(
            pretrained_dataset_id,
            int(args.tta_dataset_id),
            pretrainer=args.pretrainer,
            pretrainer_config=args.pretrainer_config,
            pretrainer_fold=pretrainer_fold,
            tta_dataset_bucket=args.tta_dataset_bucket,
        )

    def run_tta(self):
        check_trainers_injected()
        parser = argparse.ArgumentParser(description="Run DG-TTA")
        parser.add_argument(
            "pretrained_dataset_id",
            help="""
                            Task ID for pretrained model.
                            Can be numeric or one of ['TS104_GIN', 'TS104_MIND', 'TS104_GIN_MIND']""",
        )
        parser.add_argument("tta_dataset_id", help="Task ID for TTA")
        parser.add_argument(
            "--pretrainer",
            help="Trainer to use for pretraining",
            default=None,
        )
        parser.add_argument(
            "--pretrainer_config",
            help="Fold ID of nnUNet model to use for pretraining",
            default="3d_fullres",
        )
        parser.add_argument(
            "--pretrainer_fold",
            help="Fold ID of nnUNet model to use for pretraining",
            default="0",
        )
        parser.add_argument("--device", help="Device to be used", default="cuda")

        args = parser.parse_args(sys.argv[2:])
        pretrained_dataset_id = (
            int(args.pretrained_dataset_id)
            if args.pretrained_dataset_id.isnumeric()
            else args.pretrained_dataset_id
        )
        pretrainer_fold = (
            int(args.pretrainer_fold)
            if args.pretrainer_fold.isnumeric()
            else args.pretrainer_fold
        )

        (
            tta_data_dir,
            plan_dir,
            results_dir,
            pretrained_dataset_name,
            tta_dataset_name,
        ) = get_tta_folders(
            pretrained_dataset_id,
            int(args.tta_dataset_id),
            args.pretrainer,
            args.pretrainer_config,
            pretrainer_fold,
        )

        now_str = datetime.now().strftime("%Y%m%d__%H_%M_%S")
        numbers = [
            int(re.search(r"[0-9]+$", str(_path))[0]) for _path in results_dir.iterdir()
        ]
        if len(numbers) == 0:
            run_no = 0
        else:
            run_no = torch.as_tensor(numbers).max().item() + 1

        run_name = f"{now_str}_{randomname.get_name()}-{run_no}"

        with open(Path(plan_dir / "tta_plan.json"), "r") as f:
            config = json.load(f)

        with open(
            Path(plan_dir) / f"{pretrained_dataset_name}_label_mapping.json", "r"
        ) as f:
            pretrained_label_mapping = json.load(f)

        with open(Path(plan_dir) / f"{tta_dataset_name}_label_mapping.json", "r") as f:
            tta_dataset_label_mapping = json.load(f)

        label_mapping = generate_label_mapping(
            pretrained_label_mapping, tta_dataset_label_mapping
        )
        modifier_fn_module = load_current_modifier_functions(plan_dir)
        device = torch.device(args.device)

        kwargs = dict(
            run_name=run_name,
            config=config,
            tta_data_dir=tta_data_dir,
            save_base_path=results_dir,
            label_mapping=label_mapping,
            modifier_fn_module=modifier_fn_module,
            device=device,
        )

        if wandb_run_is_available():
            wandb_run("DG-TTA", tta_main, **kwargs)
            sys.exit(0)

        tta_main(**kwargs)


def main():
    if len(sys.argv) == 1 or sys.argv[1] in ["--help", "-h"]:
        check_dga_root_is_set(soft_check=True)
    else:
        check_dga_root_is_set()
    DGTTAProgram()


if __name__ == "__main__":
    main()



import torch
import torch.nn as nn

from nnunetv2.training.nnUNetTrainer.nnUNetTrainer import nnUNetTrainer
from nnunetv2.utilities.plans_handling.plans_handler import (
    PlansManager,
    ConfigurationManager,
)
from nnunetv2.utilities.get_network_from_plans import get_network_from_plans

from dg_tta.gin import gin_hook
from dg_tta.mind import mind_hook
from dg_tta.utils import enable_internal_augmentation


class nnUNetTrainer_GIN_MIND(nnUNetTrainer):
    def __init__(
        self,
        plans: dict,
        configuration: str,
        fold: int,
        dataset_json: dict,
        unpack_dataset: bool = True,
        device: torch.device = torch.device("cuda"),
    ):
        super().__init__(
            plans, configuration, fold, dataset_json, unpack_dataset, device
        )
        self.num_epochs = 1000

    @staticmethod
    def build_network_architecture(
        plans_manager: PlansManager,
        dataset_json,
        configuration_manager: ConfigurationManager,
        num_input_channels,
        enable_deep_supervision: bool = True,
    ) -> nn.Module:
        num_input_channels = 12
        network = get_network_from_plans(
            plans_manager,
            dataset_json,
            configuration_manager,
            num_input_channels,
            deep_supervision=enable_deep_supervision,
        )

        enable_internal_augmentation()
        gin_hook_handle = network.register_forward_pre_hook(gin_hook)
        mind_hook_handle = network.register_forward_pre_hook(mind_hook)

        return network
